=begin
= rd2html-lib.rb
=end

require "cgi"
require "rd/rdvisitor"

module RD
  class RD2HTMLVisitor < RDVisitor
    include AutoLabel
    include MethodParse
    
    # must-have constants
    OUTPUT_SUFFIX = "html"
    INCLUDE_SUFFIX = ["html"]
    
    METACHAR = { "<" => "&lt;", ">" => "&gt;", "&" => "&amp;" }

    attr(:css, true)
    attr(:charset, true)
    alias charcode charset
    alias charcode= charset=
    attr(:lang, true)
    attr(:title, true)
    attr(:html_link_rel, nil)
    attr(:html_link_rev, nil)
    attr(:use_old_anchor, true)
    # output external Label file.
    attr(:output_rbl, true)

    def initialize
      @css = nil
      @charset = nil
      @lang = nil
      @title = nil
      @html_link_rel = {}
      @html_link_rev = {}

      @footnotes = []
      @index = {}

      #      @use_old_anchor = nil
      @use_old_anchor = true # MUST -> nil
      @output_rbl = nil
      super
    end
    
    def visit(tree)
      prepare_labels(tree, "label:")
      tmp = super(tree)
      make_rbl_file(@filename) if @output_rbl and @filename
      tmp
    end

    def apply_to_DocumentElement(element, content)
      content = content.join("")
      index = make_index()
      foottext = make_foottext()
      title = @title || @filename || ARGF.filename || "Untitled"
      ret = %Q[]
      if @charset
	ret << %Q[<?xml version="1.0" encoding="#{@charset}" ?>\n]
      else 
	ret << %Q[<?xml version="1.0" ?>\n]
      end
      ret << %Q[<!DOCTYPE html \n]
      ret << %Q[  PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"\n]
      ret << %Q[  "DTD/xhtml1-strict.dtd">\n]
      if @lang
	ret << %Q[<html xmlns="http://www.w3.org/1999/xhtml" ] +
	  %Q[lang="#{@lang}" xml:lang="#{@lang}">\n]
      else
	ret << %Q[<html xmlns="http://www.w3.org/1999/xhtml">\n]
      end
      ret << %Q[<head>\n<title>#{title}</title>\n]
      if @charset
	ret << %Q[<meta http-equiv="Content-type" ] +
	  %Q[content="text/html; charset=#{@charset}" />\n]
      end
      if @css
	ret << %Q[<link href="#{@css}" type="text/css" rel="stylesheet" />\n]
      end # "

      @html_link_rel.each do |rel, href|
	ret << %Q[<link rel="#{rel}" href="#{href}" />\n]
      end
      @html_link_rev.each do |rev, href|
	ret << %Q[<link rev="#{rev}" href="#{href}" />\n]
      end

      ret << %Q[</head>\n]
#      ret << %Q[<BODY>\n#{content}#{index}#{foottext}\n</BODY>\n</HTML>]
      ret << %Q[<body>\n#{content}#{foottext}\n</body>\n</html>\n]
      ret
    end
    
    def apply_to_Headline(element, title)
      anchor = refer(element)
      label = hyphen_escape(element.label)
      %Q[<h#{element.level}><a name="#{anchor}" id="#{anchor}">#{title}</a>] +
      %Q[</h#{element.level}><!-- RDLabel: "#{label}" -->\n]
    end

    # RDVisitor#apply_to_Include 

    def apply_to_TextBlock(element, content)
      content = content.join("")
      if (element.parent.is_a?(ItemListItem) or
	  element.parent.is_a?(EnumListItem)) and
	  consist_of_one_textblock?(element.parent)
	content
      else
	%Q[<p>\n#{content}</p>\n]
      end
    end

    def apply_to_Verbatim(element)
      content = []
      element.each_line do |i|
	content.push(apply_to_String(i))
      end
      %Q[<pre>\n#{content.join("")}</pre>\n]
    end
  
    def apply_to_ItemList(element, items)
      %Q[<ul>\n#{items.join("")}</ul>\n]
    end
  
    def apply_to_EnumList(element, items)
      %Q[<ol>\n#{items.join("")}</ol>\n]
    end
    
    def apply_to_DescList(element, items)
      %Q[<dl>\n#{items.join("\n")}\n</dl>\n]
    end

    def apply_to_MethodList(element, items)
      %Q[<dl>\n#{items.join("\n")}\n</dl>\n]
    end
    
    def apply_to_ItemListItem(element, content)
      %Q[<li>#{content.join("\n")}</li>]
    end
    
    def apply_to_EnumListItem(element, content)
      %Q[<li>#{content.join("\n")}</li>]
    end

    def consist_of_one_textblock?(listitem)
      listitem.content.size == 1 and listitem.content[0].is_a?(TextBlock)
    end
    private :consist_of_one_textblock?
    
    def apply_to_DescListItem(element, term, description)
      anchor = refer(element)
      label = hyphen_escape(element.label)
      if description.empty?
	%Q[<dt><a name="#{anchor}" id="#{anchor}">#{term}</a></dt>] +
	%Q[<!-- RDLabel: "#{label}" -->]
      else
        %Q[<dt><a name="#{anchor}" id="#{anchor}">#{term}</a></dt>] +
        %Q[<!-- RDLabel: "#{label}" -->\n] +
        %Q[<dd>\n#{description.join("\n")}</dd>]
      end
    end

    def apply_to_MethodListItem(element, term, description)
      term = parse_method(term)  # maybe: term -> element.term
      anchor = refer(element)
      label = hyphen_escape(element.label)
      if description.empty?
	%Q[<dt><a name="#{anchor}" id="#{anchor}"><code>#{term}] +
        %Q[</code></a><!-- RDLabel: "#{label}" -->]
      else
        %Q[<dt><a name="#{anchor}" id="#{anchor}"><code>#{term}] +
	%Q[</code></a></dt><!-- RDLabel: "#{label}" -->\n] +
	%Q[<dd>\n#{description.join("\n")}</dd>]
      end
    end
  
    def apply_to_StringElement(element)
      apply_to_String(element.content)
    end
    
    def apply_to_Emphasis(element, content)
      %Q[<em>#{content.join("")}</em>]
    end
  
    def apply_to_Code(element, content)
      %Q[<code>#{content.join("")}</code>]
    end
  
    def apply_to_Var(element, content)
      %Q[<var>#{content.join("")}</var>]
    end
  
    def apply_to_Keyboard(element, content)
      %Q[<kbd>#{content.join("")}</kbd>]
    end
  
    def apply_to_Index(element, content)
      tmp = []
      element.each do |i|
	tmp.push(i) if i.is_a?(String)
      end
      key = meta_char_escape(tmp.join(""))
      if @index.has_key?(key)
	# warning?
	%Q[<!-- Index, but conflict -->#{content.join("")}<!-- Index end -->]
      else
	num = @index[key] = @index.size
	%Q[<a name="index:#{num}" id="index:#{num}">#{content.join("")}</a>]
      end
    end

    def apply_to_Reference(element, content)
      case element.label
      when Reference::URL
	apply_to_RefToURL(element, content)
      when Reference::RDLabel
	if element.label.filename
	  apply_to_RefToOtherFile(element, content)
	else
	  apply_to_RefToElement(element, content)
	end
      end
    end

    def apply_to_RefToElement(element, content)
      content = content.join("")
      if anchor = refer(element)
	content = content.sub(/^function#/, "")
	%Q[<a href="\##{anchor}">#{content}</a>]
      else
	# warning?
	label = hyphen_escape(element.to_label)
	%Q[<!-- Reference, RDLabel "#{label}" doesn't exist -->] +
	  %Q[<em class="label-not-found">#{content}</em><!-- Reference end -->]
	#' 
      end
    end

    def apply_to_RefToOtherFile(element, content)
      content = content.join("")
      filename = element.label.filename.sub(/\.(rd|rb)(\.\w+)?$/, "." +
					    OUTPUT_SUFFIX)
      anchor = refer_external(element)
      if anchor
	%Q[<a href="#{filename}\##{anchor}">#{content}</a>]
      else
	%Q[<a href="#{filename}">#{content}</a>]
      end
    end
  
    def apply_to_RefToURL(element, content)
      %Q[<a href="#{meta_char_escape(element.label.url)}">] +
	%Q[#{content.join("")}</a>]
    end
    
    def apply_to_Footnote(element, content)
      @footnotes.push(content)
      num = @footnotes.size
      %Q[<a name="footnote:#{num}" id="footnote:#{num}" ] +
      %Q[ href="#foottext:#{num}"><sup><small>] +
      %Q[*#{num}</small></sup></a>]
    end
    
    def apply_to_Verb(element)
      content = sp2nbsp(apply_to_String(element.content))
      %Q[#{content}]
    end

    def sp2nbsp(str)
      str.gsub(/\s/, "&nbsp;")
    end
    private :sp2nbsp
    
    def apply_to_String(element)
      meta_char_escape(element)
    end

    def prepare_labels(tree, prefix = "label:")
      if @use_old_anchor
	super
      else
	prepare_labels_with_url_escape(tree, prefix)
      end
    end
    private :prepare_labels
      
    def prepare_labels_with_url_escape(tree, prefix = "label:")
      @__labels__ = {}
      @__external_labels__ = {}
      @__tree__ = tree

      # preprocessing for Reference
      headlines = tree.find_all do |i|
	i.is_a?(Headline)
      end
      desclists = tree.find_all do |i|
	i.is_a?(DescList)
      end
      methodlists = tree.find_all do |i|
	i.is_a?(MethodList)
      end
      
      headlines.each do |i|
	@__labels__[i.label] ||= prefix + CGI.escape(i.label)
      end
      desclists.each do |i|
	i.each_item do |j|
	  @__labels__[j.label] ||= prefix + CGI.escape(j.label)
	end
      end
      methodlists.each do |i|
	i.each_item do |j|
	  if self.respond_to?(:make_mindex_label, true)
	    @__labels__[j.label] ||= make_mindex_label(j)
	  else
	    @__labels__[j.label] ||= prefix + CGI.escape(j.label)
	  end
	end
      end
      @__labels__
    end
    private :prepare_labels_with_url_escape

    def refer_external(element)
      if @use_old_anchor
	super
      else
	refer_external_with_url_escape(element)
      end
    end

    def refer_external_with_url_escape(element)
      case element
      when Reference
	if element.label.element_label
	  CGI.escape(element.to_label)
	else
	  nil
	end
      when Reference::Label
	if element.element_label
	  CGI.escape(element.to_label)
	else
	  nil
	end
      end
    end
    
    def parse_method(method)
      klass, kind, method, args = MethodParse.analize_method(method)
      
      if kind == :function
	klass = kind = nil
      else
	kind = MethodParse.kind2str(kind)
      end
      
      args.gsub!(/\w+/, '<var>\\&</var>')

      case method
      when "[]"
	args.strip!
	args.sub!(/^\((.*)\)$/, '\\1')
	"#{klass}#{kind}[#{args}]"
      when "[]="
	args.tr!(' ', '')
	args.sub!(/^\((.*)\)$/, '\\1')
	ary = args.split(/,/)

	case ary.length
	when 1
	  val = '<var>val</var>'
	when 2
	  args, val = *ary
	when 3
	  args, val = ary[0, 2].join(', '), ary[2]
	end

	"#{klass}#{kind}[#{args}] = #{val}"
      else
	"#{klass}#{kind}#{method}#{args}"
      end
    end
    private :parse_method

    def meta_char_escape(str)
      str.gsub(/[<>&]/) {
	METACHAR[$&]
      }
    end
    private :meta_char_escape

    def hyphen_escape(str)
      str.gsub(/--/, "&shy;&shy;")
    end
    
    def make_index
      unless @index.empty?
	tmp = []
	@index.keys.sort.each do |entry|
	  num = @index[entry]
	  tmp.push(%[<li><a href="#index:#{num}">#{entry}</A>])
	end
	tmp = tmp.join("\n")
	%[\n<hr />\n<ul>\n#{tmp}\n</ul>]
      end
    end
    private :make_index
    
    def make_foottext
      if @footnotes[0]
	ret = "\n<hr />\n<p>\n"
	num = 0
	while i = @footnotes.shift
	  num += 1
	  ret << %Q[<a name="foottext:#{num}"] +
	    %Q[ id="foottext:#{num}" href="#footnote:#{num}"><sup>] +
	    %Q[<small>*#{num}</small></sup></a>] +
	    %Q[<small>#{i}</small><br />\n]
	end
	ret << "</p>"
	ret
      else
	""
      end
    end
    private :make_foottext
    
  end # RD2HTMLVisitor
end # RD

$Visitor_Class = RD::RD2HTMLVisitor
$RD2_Sub_OptionParser = "rd/rd2html-opt"

=begin
== script info.
 RD to HTML translate library for rdfmt.rb
 $Id: rd2html-lib.rb,v 1.31.4.9 2001/01/27 19:39:03 toshirok Exp $

=end



